#include "qui/motif_editor/MotifEditorDelegate.hpp"

#include <Qt>
#include <QCheckBox>
#include <QColorDialog>
#include <QEvent>
#include <QKeyEvent>
#include <QLineEdit>
#include <QPainter>
#include <QRegExp>
#include <QRegExpValidator>

MotifEditorDelegate::MotifEditorDelegate(QWidget *parent) :
  QItemDelegate(parent)
{
}

QWidget *MotifEditorDelegate::createEditor(
  QWidget *parent, 
  const QStyleOptionViewItem & /* option */,
  const QModelIndex &index) const
{
  static const QRegExp no_quotes("[^\"]*");
  // grabbed from Alphabet::nucleic_alphabet
  static const QRegExp iupac("[AaCcGgTtUuRrYyMmKkSsWwBbDdHhVvNn-~.?]*");
  QRegExpValidator *regex_validator;
  QLineEdit *widget = 0;
  switch(index.column()) {
    // case MotifModel::EnabledCell is handled in editorEvent
    // case MotifModel::ColorCell:
    case MotifModel::NameCell:
      widget = new QLineEdit(parent);
      widget->installEventFilter(const_cast<MotifEditorDelegate*>(this));
      regex_validator = new QRegExpValidator(no_quotes, parent);
      widget->setValidator(regex_validator);
      break;
    case MotifModel::SequenceCell:
      widget = new QLineEdit(parent);
      widget->installEventFilter(const_cast<MotifEditorDelegate*>(this));
      regex_validator = new QRegExpValidator(iupac, parent);
      widget->setValidator(regex_validator);
      break;
    default:
      break;
  }
  return widget;
}

//! check to see if our event is something that should "activate" our cell
static bool is_cell_activated(QEvent *event)
{
  // if a mouse releases on us we can toggle
  if (event->type() == QEvent::MouseButtonRelease) {
    return true;
  // if we're a keypress
  } else if (event->type() == QEvent::KeyPress) {
    // try to convert to a KeyEvent
    QKeyEvent *key_event = dynamic_cast<QKeyEvent *>(event);
    // if we've gotten our key event, did the use hit the enter/return key?
    if (key_event) {
      if (key_event->key() == Qt::Key_Enter or key_event->key() == Qt::Key_Return) {
        return true;
      }
    }
  } 
  // all other cases don't toggle
  return false;  
}
bool MotifEditorDelegate::editorEvent(
       QEvent *event, 
       QAbstractItemModel *model, 
       const QStyleOptionViewItem &option, 
       const QModelIndex & index )
{
  switch (index.column()) {
    case MotifModel::EnabledCell:
      if (is_cell_activated(event)) {
        QVariant value = index.model()->data(index, Qt::DisplayRole);
        value = not value.toBool();
        model->setData(index, value);
        return true;  
      }
      break;
    case MotifModel::ColorCell:
      if (is_cell_activated(event)) {
        QVariant value = index.model()->data(index, Qt::DisplayRole);
        QColor old_color = value.value<QColor>();
        QColor new_color = QColorDialog::getColor(old_color);
        // color is not valid if user hit the cancel button on the color dialog
        if (new_color.isValid()) {
          model->setData(index, QVariant(new_color));
        } else {
          model->setData(index, QVariant(old_color));
        }
        return true;
      }
      break;
    default:
      break;
  }            
  return this->QItemDelegate::editorEvent(event, model, option, index);
} 

void MotifEditorDelegate::setEditorData(
  QWidget *editor, 
  const QModelIndex &index) const
{
  QVariant value = index.model()->data(index, Qt::DisplayRole);
  switch(index.column()) {
    // toggling MotifModel::EnabledCell is handled by editorEvent
    // case MotifModel::ColorCell is handled by paint and editorEvent

    // both of these should be text boxes
    case MotifModel::NameCell:
    case MotifModel::SequenceCell:
      static_cast<QLineEdit*>(editor)->setText(value.toString());
      break;
    default:
      break;
  }
}

void MotifEditorDelegate::setModelData(
  QWidget *editor, 
  QAbstractItemModel *model,
  const QModelIndex &index) const
{
  QVariant value;
  switch(index.column()) {
    // both of these should be text boxes
    case MotifModel::NameCell:
    case MotifModel::SequenceCell:
      value = static_cast<QLineEdit*>(editor)->text();
      model->setData(index, value);  
      break;
    default:
      break;
  }  
}

struct painter_save_restore {
  painter_save_restore(QPainter *p) { painter = p; painter->save(); }
  ~painter_save_restore() { painter->restore(); }
  QPainter *painter;
};

static void drawHighlightIfNeeded(QPainter *painter, const QStyleOptionViewItem &option)
{
    if (option.state & QStyle::State_Selected) { 
        painter->setBrush(option.palette.highlight());
        painter->drawRect(option.rect);         
    } 
}

void MotifEditorDelegate::paint(
  QPainter * painter, 
  const QStyleOptionViewItem & option, 
  const QModelIndex & index ) const
{
  painter_save_restore saved(painter);
  
  QVariant value = index.model()->data(index, Qt::DisplayRole);
  switch(index.column()) {
    case MotifModel::EnabledCell:
    {
      painter->setPen(Qt::NoPen);
      drawHighlightIfNeeded(painter, option);
      Qt::CheckState state = (value.toBool()) ? Qt::Checked : Qt::Unchecked;
      this->QItemDelegate::drawCheck(painter, option, option.rect, state);
    }
    break;
    case MotifModel::ColorCell:
    {
      painter->setPen(Qt::NoPen);
      drawHighlightIfNeeded(painter, option);       
      QColor color(value.value<QColor>());
      painter->setBrush(QBrush(color));
      painter->drawRoundRect(option.rect, 60, 60);
    }  
    break;
    default:
      this->QItemDelegate::paint(painter, option, index);
    break;
  }
}

void MotifEditorDelegate::updateEditorGeometry(
  QWidget *editor,
  const QStyleOptionViewItem &option, 
  const QModelIndex &index) const
{
  editor->setGeometry(option.rect);
}
  
